/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  HTTP2CSI 
*
* @FILE     http2csi
*
* @ABSTRACT
* Apache Common Http2csi  handler
*
* This file contains:
*
* @INCLUDE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

/* Apache */
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_core.h"
#include "fnmatch.h"

/* expat XML parser */
#include "expat.h"

/* MQ-Series include files */
#include <cmqc.h>

#include "gcn_uptyp.h"
#include "gcn_upcon.h"
#include "gcn_upext.h"
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upext.h"
#include "loglo_uptyp.h"
#include "evr_upmac.h"
#include "evr_upext.h"

#include "com_upext.h"
#include "tools_upext.h"

#include "http2csi_mqm_upext.h"

#include "http2csi_prcon.h"
#include "http2csi_prmes.h"
#include "http2csi_prtyp.h"
#include "http2csi_upext.h"

module MODULE_VAR_EXPORT Http2csiModule;

static HTTP2CSICONTEXT sContext = {HTTP2CSICONTEXT_DEFAULT};

static void h2c_rdebug(request_rec *r, char *format, ...)
{
    char work_buf[1024*10];
    va_list argptr;

    va_start(argptr, format);
    vsprintf(work_buf, format, argptr);
    va_end(argptr);

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, r, work_buf);

    EVR_TRACELOG(EVR_DEBUG, (H2C_MSG_DEBUG_LABEL, work_buf));
}
static void h2c_sdebug(server_rec *s, char *format, ...)
{
    char work_buf[1024*10];
    va_list argptr;

    va_start(argptr, format);
    vsprintf(work_buf, format, argptr);
    va_end(argptr);

    ap_log_error(APLOG_MARK, APLOG_DEBUG, s, work_buf);

    EVR_TRACELOG(EVR_DEBUG, (H2C_MSG_DEBUG_LABEL, work_buf));
}
    
static void InitContext()
{
    sContext.szArgReq             = NULL;
    sContext.szArgAction          = NULL;
    sContext.szArgNbserver        = NULL;
    sContext.szApp                = NULL;
    sContext.szMode               = NULL;
    sContext.szSectionName        = NULL;
    sContext.szMessage            = NULL;
    sContext.lTimeout             = 0L;
    sContext.lExpiry              = 0L;
    sContext.buffer               = NULL;
    sContext.lBufferLength        = 0L;
    sContext.lCntElemCsiRequest   = 0L;
    sContext.lCntElemCsiCallReq   = 0L;
    sContext.lCntElemAdminService = 0L;
    sContext.AdmXmlnsCcncsi       = NULL;
    sContext.AdmCommand           = NULL;
    sContext.AdmNbserver          = NULL;
    sContext.ReqXmlnsCcncsi       = NULL;
    sContext.ReqName              = NULL;
    sContext.CallService          = NULL;
    sContext.CallMsgtype          = NULL;
    sContext.CallHostformat       = NULL;
    sContext.CallCodepage         = NULL;
    sContext.CallTimeout          = NULL;
    sContext.CallEncoding         = NULL;
        
    sContext.szReqQueueName       = NULL;
    sContext.szRespQueueName      = NULL;
    sContext.szReqRptQueueName    = NULL;
    sContext.szRespRptQueueName   = NULL;
}

/*----------------------------------------------------------------------*/
/*                               HTTP out                               */
/*----------------------------------------------------------------------*/

static long PutHttpResponse(request_rec *r)
{
    r->content_type = "text/html";
    ap_send_http_header(r);

    sContext.buffer[sContext.lBufferLength] = '\0';
    ap_rputs(sContext.buffer, r);
}


/*----------------------------------------------------------------------*/
/*                               HTTP in                                */
/*----------------------------------------------------------------------*/

static long ParseURI(request_rec *r)
{
    table *args = NULL;
    char *svc = NULL;

    h2c_rdebug (r, "r->unparsed_uri        <%s>", r->unparsed_uri     );
    h2c_rdebug (r, "r->uri                 <%s>", r->uri              );
    h2c_rdebug (r, "r->filename            <%s>", r->filename         );
    h2c_rdebug (r, "r->args                <%s>", r->args             );

    /* parse URI  (APPD part of URL + param req & action) */
    if (strlen(r->path_info) < 2)
        return HTTP_BAD_REQUEST;
    sContext.szApp = ap_pstrdup(r->pool, r->path_info + 1); /* r->path_info + 1 : remove leading slash (/) */
    h2c_rdebug (r, "sContext.szApp         <%s>", sContext.szApp      );

    args = ArgumentsParse(r);
    if ( ! (sContext.szArgReq    = ap_table_get(args, "req"   )) )
        return HTTP_BAD_REQUEST;
    h2c_rdebug (r, "sContext.szArgReq      <%s>", sContext.szArgReq   );

    if ( ! (sContext.szArgAction = ap_table_get(args, "action")) )
        return HTTP_BAD_REQUEST;
    h2c_rdebug (r, "sContext.szArgAction   <%s>", sContext.szArgAction);

    sContext.szArgNbserver = ap_table_get(args, "nbserver");
    if ( sContext.szArgNbserver )
        h2c_rdebug (r, "sContext.szArgNbserver   <%s>", sContext.szArgNbserver);

    /* parse URI (mode) */
    svc = ap_getword(r->pool, (const char**) &(r->uri), '/');
    svc = ap_getword(r->pool, (const char**) &(r->uri), '/');
    h2c_rdebug (r, "svc <%s>", svc);
    if      (( strcmp(svc, "HTTP2CSI-LCT") == 0 ) || ( strcmp(svc, "HTTP2CSI-ADMIN-LCT") == 0 ))
    {
        sContext.szMode = ap_pstrdup(r->pool, ".LCT");
        sContext.szSectionName = ap_psprintf(r->pool, "%s-LCT", sContext.szApp);
    }
    else if (( strcmp(svc, "HTTP2CSI-LST") == 0 ) || ( strcmp(svc, "HTTP2CSI-ADMIN-LST") == 0 ))
    {
        sContext.szMode = ap_pstrdup(r->pool, ".LST");
        sContext.szSectionName = ap_psprintf(r->pool, "%s-LST", sContext.szApp);
    }
    else if (( strcmp(svc, "HTTP2CSI-RIT") == 0 ) || ( strcmp(svc, "HTTP2CSI-ADMIN-RIT") == 0 ))
    {
        sContext.szMode = ap_pstrdup(r->pool, ".RIT");
        sContext.szSectionName = ap_psprintf(r->pool, "%s-RIT", sContext.szApp);
    }
    else if (( strcmp(svc, "HTTP2CSI-RCT") == 0 ) || ( strcmp(svc, "HTTP2CSI-ADMIN-RCT") == 0 ))
    {
        sContext.szMode = ap_pstrdup(r->pool, ".RCT");
        sContext.szSectionName = ap_psprintf(r->pool, "%s-RCT", sContext.szApp);
    }
    else if (( strcmp(svc, "HTTP2CSI")     == 0 ) || ( strcmp(svc, "HTTP2CSI-ADMIN") == 0 ))
    {
        sContext.szMode = ap_pstrdup(r->pool, "");
        sContext.szSectionName = ap_psprintf(r->pool, "%s", sContext.szApp);
    }
    else
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "Invalid mode <%s>", svc);
        EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INVALID_MODE, svc));
        return HTTP_BAD_REQUEST;
    }
    h2c_rdebug (r, "sContext.szMode        <%s>", sContext.szMode);
    h2c_rdebug (r, "sContext.szSectionName <%s>", sContext.szSectionName);

    /* build queue names */
    if (!(sContext.szReqQueueName  = ap_psprintf(r->pool
                                                ,"%s%s%s.%s"
                                                ,HTTP2CSI_REQUEST_QUEUE_PFX
                                                ,sContext.szMode
                                                ,HTTP2CSI_REQUEST_QUEUE_SFX
                                                ,sContext.szApp
                                                )
         )
       )
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "Insuficient memory in %s line %d", __FILE__, __LINE__);
        EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INSUFICIENT_MEMORY, __FILE__, __LINE__));
        return HTTP_INSUFFICIENT_STORAGE;
    }
    else
    {
        h2c_rdebug(r, "sContext.szReqQueueName <%s>", sContext.szReqQueueName);
    }

    if ( ! (sContext.szRespQueueName = ap_psprintf(r->pool, "%s%s%s.%s", HTTP2CSI_RESPONSE_QUEUE_PFX, sContext.szMode, HTTP2CSI_RESPONSE_QUEUE_SFX, sContext.szApp)))
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "Insuficient memory in %s line %d", __FILE__, __LINE__);
        EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INSUFICIENT_MEMORY, __FILE__, __LINE__));
        return HTTP_INSUFFICIENT_STORAGE;
    }
    else
    {
        h2c_rdebug(r, "sContext.szRespQueueName <%s>", sContext.szRespQueueName);
    }
    
    if ( ! (sContext.szReqRptQueueName  = ap_psprintf(r->pool, "%s%s%s.%s", HTTP2CSI_REQUEST_RPT_QUEUE_PFX , sContext.szMode, HTTP2CSI_REQUEST_RPT_QUEUE_SFX , sContext.szApp)))
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "Insuficient memory in %s line %d", __FILE__, __LINE__);
        EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INSUFICIENT_MEMORY, __FILE__, __LINE__));
        return HTTP_INSUFFICIENT_STORAGE;
    }
    else
    {
        h2c_rdebug(r, "sContext.szReqRptQueueName <%s>", sContext.szReqRptQueueName);
    }

    if ( ! (sContext.szRespRptQueueName = ap_psprintf(r->pool, "%s%s%s.%s", HTTP2CSI_RESPONSE_RPT_QUEUE_PFX, sContext.szMode, HTTP2CSI_RESPONSE_RPT_QUEUE_SFX, sContext.szApp)))
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "Insuficient memory in %s line %d", __FILE__, __LINE__);
        EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INSUFICIENT_MEMORY, __FILE__, __LINE__));
        return HTTP_INSUFFICIENT_STORAGE;
    }
    else
    {
        h2c_rdebug(r, "sContext.szRespRptQueueName <%s>", sContext.szRespRptQueueName);
    }
    
    return OK;
}

static int LogAppdContext(void *data, const char *key, const char *val)
{
    server_rec *s = (server_rec *)data;
    HTTP2CSIAPPDCONTEXT *psAppdContext = (HTTP2CSIAPPDCONTEXT *)val;
    h2c_sdebug(s, "LogAppdContext <%s> lRequestTimeoutDefault <%d>", key, psAppdContext->lRequestTimeoutDefault);
    h2c_sdebug(s, "LogAppdContext <%s> lRequestTimeoutMin     <%d>", key, psAppdContext->lRequestTimeoutMin);
    h2c_sdebug(s, "LogAppdContext <%s> lRequestTimeoutMax     <%d>", key, psAppdContext->lRequestTimeoutMax);
    h2c_sdebug(s, "LogAppdContext <%s> lAdminRequestTimeout   <%d>", key, psAppdContext->lAdminRequestTimeout);
    h2c_sdebug(s, "LogAppdContext <%s> lAdminRequestExpiry    <%d>", key, psAppdContext->lAdminRequestExpiry);
    return TRUE;
}
static long LogAllAppdContexts(server_rec *s)
{
    h2c_sdebug(s, "LogAllAppdContexts");
    ap_table_do(LogAppdContext, s, sContext.ptAppdContext, NULL);
    return OK;
}

static HTTP2CSIAPPDCONTEXT *GetAppdContext()
{
    HTTP2CSIAPPDCONTEXT *psAppdContext;
    char    *szAppKey;

    psAppdContext = (HTTP2CSIAPPDCONTEXT *) ap_table_get(sContext.ptAppdContext, sContext.szApp);
    
    if ( psAppdContext == NULL )
    {
        psAppdContext = malloc(sizeof(HTTP2CSIAPPDCONTEXT));
        szAppKey      = strdup(sContext.szApp);

        psAppdContext->lRequestTimeoutDefault = (CSILONG) GetPrivateProfileInt(sContext.szSectionName
                                                                              ,H2C_CFG_REQUESTTIMEOUTDEFAULT
                                                                              ,sContext.lRequestTimeoutDefault
                                                                              ,sContext.szConfFileName
                                                                              );
        psAppdContext->lRequestTimeoutMin     = (CSILONG) GetPrivateProfileInt(sContext.szSectionName
                                                                              ,H2C_CFG_REQUESTTIMEOUTMIN
                                                                              ,sContext.lRequestTimeoutMin
                                                                              ,sContext.szConfFileName
                                                                              );
        psAppdContext->lRequestTimeoutMax     = (CSILONG) GetPrivateProfileInt(sContext.szSectionName
                                                                              ,H2C_CFG_REQUESTTIMEOUTMAX
                                                                              ,sContext.lRequestTimeoutMax
                                                                              ,sContext.szConfFileName
                                                                              );
        psAppdContext->lAdminRequestTimeout   = (CSILONG) GetPrivateProfileInt(sContext.szSectionName
                                                                              ,H2C_CFG_ADMINREQUESTTIMEOUT
                                                                              ,sContext.lAdminRequestTimeout
                                                                              ,sContext.szConfFileName
                                                                              );
        psAppdContext->lAdminRequestExpiry    = (CSILONG) GetPrivateProfileInt(sContext.szSectionName
                                                                              ,H2C_CFG_ADMINREQUESTEXPIRY
                                                                              ,sContext.lAdminRequestExpiry
                                                                              ,sContext.szConfFileName
                                                                              );

        ap_table_setn(sContext.ptAppdContext, szAppKey, (char *)psAppdContext);
    }

    return psAppdContext;
}

/* get POSTed message */
static long ReadMessageBody(request_rec *r, char **szMessage)
{
    long rc = OK;
    h2c_rdebug(r, "ReadMessageBody START");

    if ( (rc = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)))
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "ap_setup_client_block rc = <%ld>", rc);
        EVR_TRACELOG(EVR_WARNING, (H2C_MSG_SETUP_CLIENT_BLOCK, rc));
        return rc;
    }

    if (ap_should_client_block(r))
    {
        char buff[HUGE_STRING_LEN];
        int rsize, len_read, rpos = 0;
        long length = r->remaining;
        h2c_rdebug(r, "ReadMessageBody length <%d>", length);

        if (! (*szMessage = ap_pcalloc(r->pool, length + 1))) return HTTP_INSUFFICIENT_STORAGE;

        ap_hard_timeout("[RequestHandler] ReadMessageBody", r);

        while ( (len_read = ap_get_client_block(r, buff, sizeof(buff))) > 0)
        {
            h2c_rdebug(r, "ReadMessageBody len_read <%d>", len_read); 
            if ((rpos + len_read) > length)
                rsize = length - rpos;
            else 
                rsize = len_read;
            memcpy((char*)*szMessage + rpos, buff, rsize);
            rpos += rsize;
        }

        ap_kill_timeout(r);
    }
    else
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "NOT ap_should_client_block");
        EVR_TRACELOG(EVR_WARNING, (H2C_MSG_SHOULD_CLIENT_BLOCK));
        rc = HTTP_BAD_REQUEST;
    }
    h2c_rdebug(r, "ReadMessageBody return", rc);

    return rc;
}


static long BuildAdminRequestXML(request_rec *r, char **szMessage)
{
    long rc = OK;
    h2c_rdebug(r, "ReadMessageBody START");

    if (! (*szMessage = ap_pcalloc(r->pool, 1024))) return HTTP_INSUFFICIENT_STORAGE;

    if ( sContext.szArgNbserver == NULL )
    {
        if (! (*szMessage = ap_psprintf(r->pool, "<?xml version=\"1.0\"?>\n\
<ccncsi:ADMIN_SERVICE xmlns:ccncsi='http://www.ccncsi.net/schema'\n\
                command='%s' />\n"
            , sContext.szArgAction))) return HTTP_INSUFFICIENT_STORAGE;
    }
    else
    {
        if (! (*szMessage = ap_psprintf(r->pool, "<?xml version=\"1.0\"?>\n\
<ccncsi:ADMIN_SERVICE xmlns:ccncsi='http://www.ccncsi.net/schema'\n\
                command='%s' nbserver='%s' />\n"
            , sContext.szArgAction
            , sContext.szArgNbserver))) return HTTP_INSUFFICIENT_STORAGE;
    }

    h2c_rdebug(r, "ReadMessageBody return", rc);
    return rc;
}

/*----------------------------------------------------------------------*/
/*                            XML Validation                            */
/*----------------------------------------------------------------------*/

static long AttrMandatory(request_rec *r, char *attValue, char *attName, char *elemName)
{
    if (attValue == NULL)
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "attribute <%s> mandatory for element <%s>", attName, elemName);
        EVR_TRACELOG(EVR_WARNING, (H2C_MSG_ATTR_MANDATORY,  attName, elemName));
        return HTTP_BAD_REQUEST;
    }
    else
    {
        return OK;
    }
}

static long CheckAttrAdmin(request_rec *r)
{
    return OK;
}

static long CheckAttrCsi(request_rec *r)
{
    long rc;
    long i;

    h2c_rdebug(r, "CheckAttr START");

    if (sContext.lCntElemCsiCallReq > 1)
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "element CSI_CALL_REQ defined %ld times", sContext.lCntElemCsiCallReq);
        EVR_TRACELOG(EVR_WARNING, (H2C_MSG_DUPLICATE_ELEMENT,  "CSI_CALL_REQ", sContext.lCntElemCsiCallReq));
        return HTTP_BAD_REQUEST;
    }
    /* mandatory attributes */
    else if ((rc = AttrMandatory(r, sContext.ReqXmlnsCcncsi, "xmlns:ccncsi", "CSI_REQUEST" )) != OK) return rc;
    else if ((rc = AttrMandatory(r, sContext.ReqName       , "name"        , "CSI_REQUEST" )) != OK) return rc;
    else if ((rc = AttrMandatory(r, sContext.CallService   , "service"     , "CSI_CALL_REQ")) != OK) return rc;
    else if ((rc = AttrMandatory(r, sContext.CallMsgtype   , "msgtype"     , "CSI_CALL_REQ")) != OK) return rc;
    else if (strcmp(sContext.ReqName, "CSI_CALL_REQ")!=0)
    {
       ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "value not allowad for att <name> in element <CSI_REQUEST>: <%s>", sContext.ReqName);
       EVR_TRACELOG(EVR_WARNING, (H2C_MSG_ATTR_VALUE, "name", "CSI_REQUEST", sContext.ReqName));
       return HTTP_BAD_REQUEST;
    }
    else if ((sContext.CallEncoding != NULL) && (strcmp(sContext.CallEncoding, "text")!=0) && (strcmp(sContext.CallEncoding, "base64")!=0))
    {
       ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "value not allowad for att <encoding> in element <CSI_CALL_REQ>: <%s>", sContext.CallEncoding);
       EVR_TRACELOG(EVR_WARNING, (H2C_MSG_ATTR_VALUE, "encoding", "CSI_CALL_REQ", sContext.CallEncoding));
       return HTTP_BAD_REQUEST;
    }
    else
    {
        for (i = strlen(sContext.CallService); i > 0; i--)
        {
            if (sContext.CallService[i] == '.')
            {
                break;
            }
        }
        if ((sContext.CallService[i] != '.') || (strcmp(&sContext.CallService[i+1], sContext.szApp)!= 0))
        {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "attribute \"service\" <%s> not compliant with URI app <%s>", sContext.CallService, sContext.szApp);
            EVR_TRACELOG(EVR_WARNING, (H2C_MSG_INVALID_SERVICE, sContext.CallService, sContext.szApp));
            return HTTP_BAD_REQUEST;
        }
        return OK;
    }
}


/*----------------------------------------------------------------------*/
/*                             XML Parsing                              */
/*----------------------------------------------------------------------*/

static void startElement(void *userData, const char *name, const char **atts)
{
    EXPAT_DATA *psData = userData;
    request_rec *r = psData->r;
    long i;

    h2c_rdebug(r, "element <%s> at depth <%d> start", name, psData->lDepth);
    switch(psData->lDepth)
    {
        case 0 :
            if (strcmp(name, "ccncsi:CSI_REQUEST")==0)
            {
                sContext.lCntElemCsiRequest += 1;
                for (i = 0; atts[i]; i += 2)
                {
                   h2c_rdebug(r, "element <%s> attribute <%s> = <%s>", name, atts[i], atts[i+1]);
                   if      (strcmp(atts[i],"xmlns:ccncsi")==0) { sContext.ReqXmlnsCcncsi = NOT_NULL; }
                   else if (strcmp(atts[i],"name"        )==0) { sContext.ReqName        = ap_pstrdup(r->pool, atts[i+1]); }
                   else
                   {
                       ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid attribute for element %s : %s",name,atts[i]);
                       EVR_TRACELOG(EVR_WARNING, (H2C_MSG_INVALID_SERVICE, name, atts[i]));
                       psData->lReturnCode = HTTP_BAD_REQUEST;
                   }
                }
            }
            else if (strcmp(name, "ccncsi:ADMIN_SERVICE")==0)
            {
                sContext.lCntElemAdminService += 1;
                for (i = 0; atts[i]; i += 2)
                {
                   h2c_rdebug(r, "element <%s> attribute <%s> = <%s>", name, atts[i], atts[i+1]);
                   if      (strcmp(atts[i],"xmlns:ccncsi")==0) { sContext.AdmXmlnsCcncsi = NOT_NULL; }
                   else if (strcmp(atts[i],"command"     )==0) { sContext.AdmCommand     = ap_pstrdup(r->pool, atts[i+1]); }
                   else if (strcmp(atts[i],"nbserver"    )==0) { sContext.AdmNbserver    = ap_pstrdup(r->pool, atts[i+1]); }
                   else
                   {
                       ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid attribute for element %s : %s",name,atts[i]);
                       EVR_TRACELOG(EVR_WARNING, (H2C_MSG_INVALID_SERVICE, name, atts[i]));
                       psData->lReturnCode = HTTP_BAD_REQUEST;
                   }
                }
            }
            else
            {
                ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid element <%s> at depth <%ld>", name, psData->lDepth);
                EVR_TRACELOG(EVR_WARNING, (H2C_MSG_INVALID_ELEMENT, name, psData->lDepth));
                psData->lReturnCode = HTTP_BAD_REQUEST;
            }
            break;
        case 1 :
            if (strcmp(name, "ccncsi:CSI_CALL_REQ")==0)
            {
                sContext.lCntElemCsiCallReq += 1;
                for (i = 0; atts[i]; i += 2)
                {
                   h2c_rdebug(r, "element <%s> attribute <%s> = <%s>", name, atts[i], atts[i+1]);
                   if      (strcmp(atts[i],"service"   )==0) { sContext.CallService    = ap_pstrdup(r->pool, atts[i+1]); }
                   else if (strcmp(atts[i],"msgtype"   )==0) { sContext.CallMsgtype    = NOT_NULL; }
                   else if (strcmp(atts[i],"hostformat")==0) { sContext.CallHostformat = NOT_NULL; }
                   else if (strcmp(atts[i],"codepage"  )==0) { sContext.CallCodepage   = NOT_NULL; }
                   else if (strcmp(atts[i],"timeout"   )==0) { sContext.CallTimeout    = ap_pstrdup(r->pool, atts[i+1]); }
                   else if (strcmp(atts[i],"encoding"  )==0) { sContext.CallEncoding   = ap_pstrdup(r->pool, atts[i+1]); }
                   else
                   {
                       ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid attribute for element %s : %s",name,atts[i]);
                       EVR_TRACELOG(EVR_WARNING, (H2C_MSG_INVALID_SERVICE, name, atts[i]));
                       psData->lReturnCode = HTTP_BAD_REQUEST;
                   }
                }
            }
            else
            {
                ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid element <%s> at depth <%ld>", name, psData->lDepth);
                EVR_TRACELOG(EVR_WARNING, (H2C_MSG_INVALID_ELEMENT, name, psData->lDepth));
                psData->lReturnCode = HTTP_BAD_REQUEST;
            }
            break;
        default :
            psData->lReturnCode = HTTP_BAD_REQUEST;
    }

    psData->lDepth += 1;
}

static void endElement(void *userData, const char *name)
{
    EXPAT_DATA *psData = userData;
    psData->lDepth -= 1;
    h2c_rdebug(psData->r, "element <%s> at depth <%d> end", name, psData->lDepth);
}

static long ParseXML(request_rec *r, const char *szMessage)
{
    EXPAT_DATA sData;
    XML_Parser parser = XML_ParserCreate(NULL);
    long lMessageLength;

    lMessageLength = strlen(sContext.szMessage);
    h2c_rdebug(r, "ParseXML <%.*s>", (lMessageLength<200)?lMessageLength:200, sContext.szMessage);

    memset (&sData, (int)'\0', sizeof(sData));
    sData.lReturnCode = OK;
    sData.r = r;
    XML_SetUserData(parser, &sData);
    XML_SetElementHandler(parser, startElement, endElement);
    /* because the buffer is complete, we don't need a loop, a single XML_Parse is enough */
    if (XML_Parse(parser, szMessage, strlen(szMessage) , TRUE) == XML_STATUS_ERROR)
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r,
	    "XML_Parse error <%s> at line <%ld>",
	    XML_ErrorString(XML_GetErrorCode(parser)),
	    XML_GetCurrentLineNumber(parser));
        EVR_TRACELOG(EVR_WARNING, (H2C_MSG_XML_PARSE, XML_ErrorString(XML_GetErrorCode(parser)), XML_GetCurrentLineNumber(parser) ));
        sData.lReturnCode = HTTP_BAD_REQUEST;
    }
    XML_ParserFree(parser);
    return sData.lReturnCode;
}


/*----------------------------------------------------------------------*/
/*                             Queues Mgmnt                             */
/*----------------------------------------------------------------------*/

static int LogOpenQueue(void *data, const char *key, const char *val)
{
    server_rec *s = (server_rec *)data;
    h2c_sdebug(s, "LogOpenQueue <%ld> <%s> <%d>", key, key, val);
    return TRUE;
}

static long LogAllOpenQueues(server_rec *s)
{
    h2c_sdebug(s, "LogAllOpenQueues <%d>", sContext.ptQueueObj);
    ap_table_do(LogOpenQueue, s, sContext.ptQueueObj, NULL);
    return OK;
}

static MQHOBJ *GetQueueHandler(request_rec *r, char *szQueueName, long lOpeningOptions)
{
    MQHOBJ *psQueueHandler;
    char *szQueueKey;
    long lReturnCode;
    long lReasonCode;
    h2c_rdebug(r, "GetQueueHandler (%s)", szQueueName);

    if (MQHC_UNUSABLE_HCONN == sContext.hQMgr )
    {
        h2c_rdebug(r, "GetQueueHandler B");
/*
        if (MQCC_OK == h2c_mqconn(""
*/
        MQCONN(""
                                 ,&(sContext.hQMgr)
                                 ,&lReturnCode     
                                 ,&lReasonCode)
        ;
        if (MQCC_OK == lReturnCode
        )
        {
            h2c_rdebug(r, "GetQueueHandler C");
            h2c_rdebug(r, "h2c_mqconn succeed");
        }
        else
        {   
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqconn failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            sContext.hQMgr = MQHC_UNUSABLE_HCONN;
            return (MQHOBJ *) 0;
        }
    }

    psQueueHandler = (MQHOBJ *) ap_table_get(sContext.ptQueueObj, szQueueName);   

    if (psQueueHandler == NULL)
    {
        MQOD       sObjDesc = {MQOD_DEFAULT};

        psQueueHandler = malloc(sizeof(MQHOBJ));
        szQueueKey     = strdup(szQueueName);
        
        strcpy(sObjDesc.ObjectName, szQueueName);
        if (MQCC_OK == h2c_mqopen(sContext.hQMgr
                                 ,&sObjDesc     
                                 ,lOpeningOptions
                                 ,psQueueHandler
                                 ,&lReturnCode            
                                 ,&lReasonCode))
        {
            h2c_rdebug(r, "h2c_mqopen succeed <%s>", szQueueName);
            ap_table_setn(sContext.ptQueueObj, szQueueKey, (char *)psQueueHandler);
        }
        else
        {
            free(psQueueHandler);
            psQueueHandler = NULL;
            free(szQueueKey);
            szQueueKey = NULL;
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqopen failed <%s> - Return %ld Reason %ld"
                                                     , szQueueName, lReturnCode, lReasonCode);
            return (MQHOBJ *) 0;
        }
    }
    h2c_rdebug(r, "GetQueueHandler psQueueHandler = <%ld>", psQueueHandler);
    LogAllOpenQueues(r->server);

    h2c_rdebug(r, "GetQueueHandler(%s) returns <%d> (%d)", szQueueName, *psQueueHandler, psQueueHandler);
    return psQueueHandler;
}

static int CloseOpenQueue(void *data, const char *key, const char *val)
{
    long lReturnCode;
    long lReasonCode;
    server_rec *s = (server_rec *)data;

    h2c_sdebug(s, "CloseOpenQueue <%s> <%d>", key, val);
    
    {
        if (MQCC_OK == h2c_mqclose(sContext.hQMgr
                                  ,(MQHOBJ *)val
                                  ,MQCO_NONE
                                  ,&lReturnCode
                                  ,&lReasonCode
                                  ))
        {
            h2c_sdebug(s, "h2c_mqclose succeed <%s>", key);
        }
        else
        {   
            ap_log_error(APLOG_MARK, APLOG_ALERT, s, "h2c_mqclose failed <%s> - Return %ld Reason %ld"
                                                     , key, lReturnCode, lReasonCode);
        }
    }
    /*
    ap_table_unset(sContext.ptQueueObj, key);
    */
    free((void *)val);
    return TRUE;
}

static long CloseAllOpenQueues(server_rec *s)
{
    long lReturnCode;
    long lReasonCode;
    ap_table_do(CloseOpenQueue, s, sContext.ptQueueObj, NULL);
    ap_clear_table(sContext.ptQueueObj);
    
    if (MQHC_UNUSABLE_HCONN != sContext.hQMgr)
    {
        if (MQCC_OK == h2c_mqdisc(&(sContext.hQMgr)
                                 ,&lReturnCode     
                                 ,&lReasonCode))
        {
            h2c_sdebug(s, "h2c_mqdisc succeed");
        }
        else
        {   
            ap_log_error(APLOG_MARK, APLOG_ALERT, s, "h2c_mqdisc failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
        }
        sContext.hQMgr = MQHC_UNUSABLE_HCONN;
    }

    return OK;
}


/*----------------------------------------------------------------------*/
/*                            Messages Mgmnt                            */
/*----------------------------------------------------------------------*/

static long SetTimeout(request_rec *r)
{
    h2c_rdebug(r, "SetTimeout");

    if (sContext.CallTimeout == NULL)
    {
        sContext.lTimeout = (sContext.psAppdContext)->lRequestTimeoutDefault;
    }
    else
    {
        sContext.lTimeout = atol(sContext.CallTimeout);
        if (sContext.lTimeout < (sContext.psAppdContext)->lRequestTimeoutMin)
            sContext.lTimeout = (sContext.psAppdContext)->lRequestTimeoutMin;
        else 
        if (sContext.lTimeout > (sContext.psAppdContext)->lRequestTimeoutMax)
            sContext.lTimeout = (sContext.psAppdContext)->lRequestTimeoutMax;
    }
    h2c_rdebug(r, "sContext.lTimeout <%d>", sContext.lTimeout);

    return OK;
}

static long PutRequestInQueue(request_rec *r, char *szMessage)
{
    long lReturnCode;
    long lReasonCode;
    long rc = OK;
    unsigned char *sModMessage = NULL;
    CSIPEERID sDef = {CSIPEERID_DEFAULT};
    long length = 0;
    
    h2c_rdebug(r, "PutRequestInQueue <%d>", *(sContext.psReqQueueObj));
   
    /* make sure we are modifying the CSI requests and not the ADMIN requests */
    if(ap_fnmatch("*req=CSI*",r->args,0) == 0)
    {
        length = sizeof(CSIPEERID) + strlen(szMessage);
	if( (ap_table_get(r->headers_in,"X-ORIG-GTW") == NULL) ||
        	(ap_table_get(r->headers_in,"X-ORIG-SITE") == NULL))
    	{
        	ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "Missing header (X-ORIG-GTW or X-ORIG-SITE) when putting request in mq");
                return HTTP_INTERNAL_SERVER_ERROR;
    	}
     
    	strcpy(sDef.SiteName,ap_table_get(r->headers_in,"X-ORIG-SITE"));
    	strcpy(sDef.GatewayName,ap_table_get(r->headers_in,"X-ORIG-GTW"));

    	if (! (sModMessage = ap_palloc(r->pool, length + 1))) return HTTP_INSUFFICIENT_STORAGE;

    	memcpy( sModMessage, &sDef, sizeof(CSIPEERID) );
    	memcpy( sModMessage + sizeof(CSIPEERID), szMessage, strlen(szMessage)+1 );
    }
    else
    {
        length = strlen(szMessage);
        if (! (sModMessage = ap_palloc(r->pool, length + 1))) return HTTP_INSUFFICIENT_STORAGE;
        memcpy( sModMessage, szMessage, strlen(szMessage)+1 );
    }


    /* Transaction begin */
    {
        MQBO sBeginOpt = {MQBO_DEFAULT};

        if (MQCC_OK != h2c_mqbegin(sContext.hQMgr
                                  ,&sBeginOpt
                                  ,&lReturnCode
                                  ,&lReasonCode
                                  )
           )
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mbegin failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }

   /* Put request in request queue */
   {
        MQMD sMsgDesc = {MQMD_DEFAULT};
        MQPMO sPutMsgOpt = {MQPMO_DEFAULT};

        memset(sContext.szMsgId, '\0', sizeof(sContext.szMsgId));
        gcn_GetUniqueId((CSICHAR *) sContext.szMsgId);
/*
        memset(sContext.szMsgId, '\0', sizeof(sContext.szMsgId));
        strcpy(sContext.szMsgId, "123");
*/
        memcpy(sMsgDesc.MsgId, sContext.szMsgId, sizeof(sMsgDesc.MsgId));

        memset(sContext.szCorrelId, '\0', sizeof(sContext.szCorrelId));
        memcpy(sContext.szCorrelId, sContext.szArgReq, strlen(sContext.szArgReq));
        memcpy(sMsgDesc.CorrelId, sContext.szCorrelId, sizeof(sMsgDesc.CorrelId));

        h2c_rdebug(r, "sMsgDesc.MsgId 2  <%s>", sMsgDesc.MsgId);
        h2c_rdebug(r, "sMsgDesc.CorrelId <%s>", sMsgDesc.CorrelId);
        {
          int i;
          char s[64]="";
          int lSize = sizeof(sMsgDesc.MsgId);

          for ( i=0; i<lSize; i++ )
          {
            sprintf(s, "%s%.2X", s, sMsgDesc.MsgId[i]);
          }
          h2c_rdebug(r, "sMsgDesc.MsgId <%s>\n", s);
        }
        sMsgDesc.Expiry = sContext.lExpiry * 10; /* in tenth of seconds */
        sMsgDesc.Report = MQRO_EXCEPTION_WITH_FULL_DATA
                        | MQRO_EXPIRATION_WITH_FULL_DATA;
        strcpy(sMsgDesc.ReplyToQ, sContext.szReqRptQueueName);

        sPutMsgOpt.Options |= MQPMO_SYNCPOINT;
        if (MQCC_OK != h2c_mqput(sContext.hQMgr
                                ,*(sContext.psReqQueueObj)
                                ,sContext.szReqQueueName
                                ,&sMsgDesc               
                                ,&sPutMsgOpt
                                ,length
                                ,sModMessage
                                ,&lReturnCode      
                                ,&lReasonCode
                                ))           
        {
            /* Rollback the MQ-Series transaction */
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqput failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            h2c_mqback(sContext.hQMgr, &lReturnCode, &lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
        
    }

    /* Committing the MQ-Series transaction */
    {
        if (MQCC_OK != h2c_mqcmit(sContext.hQMgr, &lReturnCode, &lReasonCode))
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqcmit failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }

    return rc;
}

static long GetResponseFromQueue(request_rec *r)
{
    long lReturnCode;
    long lReasonCode;
    long rc = OK;

    h2c_rdebug(r, "GetResponseFromQueue");

    /* Transaction begin */
    {
        MQBO sBeginOpt = {MQBO_DEFAULT};

        if (MQCC_OK != h2c_mqbegin(sContext.hQMgr
                                  ,&sBeginOpt
                                  ,&lReturnCode
                                  ,&lReasonCode
                                  )
           )
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mbegin failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }

   /* Get response from response queue */
   {
        MQMD sMsgDesc = {MQMD_DEFAULT};
        MQGMO sGetMsgOpt = {MQGMO_DEFAULT};
        MQLONG lDataLength;

        memcpy(sMsgDesc.MsgId   , sContext.szMsgId,    sizeof(sMsgDesc.MsgId));
        memcpy(sMsgDesc.CorrelId, sContext.szCorrelId, sizeof(sMsgDesc.CorrelId));

        h2c_rdebug(r, "sMsgDesc.MsgId 2  <%s>", sMsgDesc.MsgId);
        h2c_rdebug(r, "sMsgDesc.CorrelId <%s>", sMsgDesc.CorrelId);
        {
          int i;
          char s[64]="";
          int lSize = sizeof(sMsgDesc.MsgId);

          for ( i=0; i<lSize; i++ )
          {
            sprintf(s, "%s%.2X", s, sMsgDesc.MsgId[i]);
          }
          h2c_rdebug(r, "sMsgDesc.MsgId <%s>\n", s);
        }

        sGetMsgOpt.MatchOptions = MQMO_MATCH_MSG_ID;
        sGetMsgOpt.MatchOptions = MQMO_NONE;
        sGetMsgOpt.Options  = (  MQGMO_NONE  );
        sGetMsgOpt.Options |= (  MQGMO_WAIT  );
        sGetMsgOpt.WaitInterval = sContext.lTimeout * 1000;
        lDataLength = 0;

        /*** browse/lock first message to get message length ***/

        h2c_mqget_datalength(sContext.hQMgr
                        ,*(sContext.psRespQueueObj)
                        ,sContext.szRespQueueName
                        ,&sMsgDesc
                        ,&sGetMsgOpt
                        ,&(sContext.lBufferLength)
                        ,&lReturnCode
                        ,&lReasonCode
                        );       
        h2c_rdebug(r, "h2c_mqget_datalength - Return %d Reason %d", lReturnCode, lReasonCode);

        if ( MQCC_OK == lReturnCode )
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqget_datalength returns 0 (MQCC_OK) and should return 1 (MQCC_WARNING) with reason 2079 (MQRC_TRUNCATED_MSG_FAILED)");
            h2c_mqback(sContext.hQMgr, &lReturnCode, &lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
        else if ( MQRC_TRUNCATED_MSG_ACCEPTED == lReasonCode )
        {
            /*** allocate data buffer (based on message length) ***/

            if (! (sContext.buffer = ap_pcalloc(r->pool, sContext.lBufferLength + 1)))
            {
                ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "unable to ap_pcalloc %ld bytes", lDataLength + 1);
                EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INSUFICIENT_MEMORY, __FILE__, __LINE__));
                h2c_mqback(sContext.hQMgr, &lReturnCode, &lReasonCode);
                return HTTP_INSUFFICIENT_STORAGE;
            }
    
            /*** get message data (and delete) ***/
    
            h2c_mqget_data(sContext.hQMgr
                        ,*(sContext.psRespQueueObj)
                        ,sContext.szRespQueueName
                        ,&sMsgDesc             
                        ,&sGetMsgOpt
                        ,(sContext.lBufferLength)
                        ,(MQBYTE*)(sContext.buffer)
                        ,&(sContext.lBufferLength)
                        ,&lReturnCode
                        ,&lReasonCode
                        ); 
            h2c_rdebug(r, "h2c_mqget_data - Return %d Reason %d", lReturnCode, lReasonCode);

            if (( MQCC_OK != lReturnCode ) && (MQRC_NO_MSG_AVAILABLE != lReasonCode))
            {   
                ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqget_data failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
                h2c_mqback(sContext.hQMgr, &lReturnCode, &lReasonCode);
                return HTTP_INTERNAL_SERVER_ERROR;
            } 
        }
        else if ( MQRC_NO_MSG_AVAILABLE == lReasonCode )
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqget_datalength timeout - Return %ld Reason %ld", lReturnCode, lReasonCode);
            h2c_mqback(sContext.hQMgr, &lReturnCode, &lReasonCode);
            return HTTP_GATEWAY_TIME_OUT;
        }
        else
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqget_datalength failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            h2c_mqback(sContext.hQMgr, &lReturnCode, &lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        h2c_rdebug(r, "sMsgDesc.MsgId    <%s>", sMsgDesc.MsgId);
        h2c_rdebug(r, "sMsgDesc.CorrelId <%s>", sMsgDesc.CorrelId);
    }

    /* Committing the MQ-Series transaction */
    {
        if (MQCC_OK != h2c_mqcmit(sContext.hQMgr, &lReturnCode, &lReasonCode))
        {
            ap_log_rerror(APLOG_MARK, APLOG_ALERT, r, "h2c_mqcmit failed - Return %ld Reason %ld", lReturnCode, lReasonCode);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }

    return rc;
}


/*----------------------------------------------------------------------*/
/*                                 Actions                              */
/*----------------------------------------------------------------------*/

static long CsiCall(request_rec *r)
{
    long rc = OK;

    if (rc = ReadMessageBody(r, &sContext.szMessage))
        return rc;

    if (sContext.szMessage == NULL)
        return HTTP_BAD_REQUEST;
    h2c_rdebug(r, "sContext.szMessage (len=%d) <%.30s>", strlen(sContext.szMessage), sContext.szMessage);

    if (rc = ParseXML(r, sContext.szMessage)) 
        return rc;

    if (rc = CheckAttrCsi(r))
        return rc;

    SetTimeout(r);
    sContext.lExpiry = sContext.lTimeout / 2;

    if ((sContext.psReqQueueObj  = GetQueueHandler(r, sContext.szReqQueueName , MQOO_OUTPUT)) == NULL)
        return HTTP_INTERNAL_SERVER_ERROR;

    if ((sContext.psRespQueueObj = GetQueueHandler(r, sContext.szRespQueueName, MQOO_BROWSE|MQOO_INPUT_SHARED)) == NULL)
        return HTTP_INTERNAL_SERVER_ERROR;
    
    /*LogAllOpenQueues(r->server);*/

    if (rc = PutRequestInQueue(r, sContext.szMessage))
        return rc;

    if (rc = GetResponseFromQueue(r))
        return rc;

    PutHttpResponse(r);

    return OK;
}

static long AdminRequest(request_rec *r)
{
    long rc = OK;

    if (rc = BuildAdminRequestXML(r, &sContext.szMessage))
        return rc;

/*
    if (rc = ReadMessageBody(r, &sContext.szMessage))
        return rc;

    if (sContext.szMessage == NULL)
        return HTTP_BAD_REQUEST;
    h2c_rdebug(r, "sContext.szMessage (len=%d) <%.30s>", strlen(sContext.szMessage), sContext.szMessage);

    if (rc = ParseXML(r, sContext.szMessage)) 
        return rc;

    if (rc = CheckAttrAdmin(r))
        return rc;
*/

    sContext.lTimeout = (sContext.psAppdContext)->lAdminRequestTimeout;
    sContext.lExpiry  = (sContext.psAppdContext)->lAdminRequestExpiry;

    if ((sContext.psReqQueueObj  = GetQueueHandler(r, sContext.szReqQueueName , MQOO_OUTPUT)) == NULL)
        return HTTP_INTERNAL_SERVER_ERROR;

    if ((sContext.psRespQueueObj = GetQueueHandler(r, sContext.szRespQueueName, MQOO_BROWSE|MQOO_INPUT_SHARED)) == NULL)
        return HTTP_INTERNAL_SERVER_ERROR;
    
    /*LogAllOpenQueues(r->server);*/

    if (rc = PutRequestInQueue(r, sContext.szMessage))
        return rc;

    if (rc = GetResponseFromQueue(r))
        return rc;

    PutHttpResponse(r);

    return OK;
}


/*----------------------------------------------------------------------*/
/*                           apache handlers                            */
/*----------------------------------------------------------------------*/

static void Http2csiChildInit(server_rec *s, pool *p)
{
    long lReturnCode;

    ap_log_error(APLOG_MARK, APLOG_DEBUG, s, "Http2csiChildInit START");
    if ((lReturnCode = EVR_OPENLOG(HTTP2CSI_NICK, HTTP2CSI_GROUP)) != CSISUCCESS)
       {
       ap_log_error(APLOG_MARK, APLOG_DEBUG, s, "Http2csiChildInit: EVR_OPENLOG failed!");
       h2c_sdebug(s, "Http2csiChildInit: EVR_OPENLOG failed!");
       }

    h2c_sdebug(s, "Http2csiChildInit START");

    /* init */
    memset (&sContext, (int)'\0', sizeof(sContext));
    sContext.ChildPool     = p;
    sContext.hQMgr         = MQHC_UNUSABLE_HCONN;
    sContext.ptQueueObj    = ap_make_table(sContext.ChildPool, 10);
    sContext.ptAppdContext = ap_make_table(sContext.ChildPool, 5);

    strcpy(sContext.szConfFileName, getenv("CCNWKD"));
    strcat(sContext.szConfFileName, "/cfg/http2csi/h2c.cfg");
    h2c_sdebug(s, "sContext.szConfFileName <%s>", sContext.szConfFileName);
    sContext.lRequestTimeoutDefault = (CSILONG) GetPrivateProfileInt(H2C_CFG_GENERAL_SECTION
                                                                    ,H2C_CFG_REQUESTTIMEOUTDEFAULT
                                                                    ,HTTP2CSI_REQUESTTIMEOUTDEFAULT
                                                                    ,sContext.szConfFileName
                                                                    );                      
    sContext.lRequestTimeoutMin     = (CSILONG) GetPrivateProfileInt(H2C_CFG_GENERAL_SECTION
                                                                    ,H2C_CFG_REQUESTTIMEOUTMIN
                                                                    ,HTTP2CSI_REQUESTTIMEOUTMIN
                                                                    ,sContext.szConfFileName
                                                                    );                      
    sContext.lRequestTimeoutMax     = (CSILONG) GetPrivateProfileInt(H2C_CFG_GENERAL_SECTION
                                                                    ,H2C_CFG_REQUESTTIMEOUTMAX
                                                                    ,HTTP2CSI_REQUESTTIMEOUTMAX
                                                                    ,sContext.szConfFileName
                                                                    );                      
    sContext.lAdminRequestTimeout   = (CSILONG) GetPrivateProfileInt(H2C_CFG_GENERAL_SECTION
                                                                    ,H2C_CFG_ADMINREQUESTTIMEOUT
                                                                    ,HTTP2CSI_ADMINREQUESTTIMEOUT
                                                                    ,sContext.szConfFileName
                                                                    );                      
    sContext.lAdminRequestExpiry    = (CSILONG) GetPrivateProfileInt(H2C_CFG_GENERAL_SECTION
                                                                    ,H2C_CFG_ADMINREQUESTEXPIRY
                                                                    ,HTTP2CSI_ADMINREQUESTEXPIRY
                                                                    ,sContext.szConfFileName
                                                                    );                      

    h2c_sdebug(s, "Http2csiChildInit END");
}

static void Http2csiChildExit(server_rec *s, pool *p)
{
    h2c_sdebug(s, "Http2csiChildExit START");

    LogAllOpenQueues(s);
    CloseAllOpenQueues(s);

    h2c_sdebug(s, "Http2csiChildExit END");
}

static int Http2csiHeaderParser(request_rec *r)
{
    long rc = OK;

    /* check if the call is a HTTP2CSI call with help from ap_fnmatch
       returns 0 when found or FNM_NOMATCH when not found.
       FNM_PATHNAME is passed to make sure that a slash in string matches
       a slash in the pattern.
    */
      
    if(ap_fnmatch("*/HTTP2CSI*/*",r->uri,0) == 0)
    {
        if(r->proxyreq) {
            ap_table_set(r->headers_in,"X-ORIG-GTW",getenv("CCN_GATEWAY_NAME"));
            ap_table_set(r->headers_in,"X-ORIG-SITE",getenv("CCN_SITE_NAME"));
        }
        else {
            if(ap_table_get(r->headers_in,"X-ORIG-GTW") == NULL){
                ap_table_set(r->headers_in,"X-ORIG-GTW",getenv("CCN_GATEWAY_NAME"));
                ap_table_set(r->headers_in,"X-ORIG-SITE",getenv("CCN_SITE_NAME"));
            }
        }    
    }
     
    return rc;
}

static int Http2csiHandler(request_rec *r) 
{
    long rc = OK;

    h2c_rdebug(r, "Http2csiHandler START");
    LogAllOpenQueues(r->server);
    InitContext();
    h2c_rdebug(r, "Http2csiHandler AFTER InitContext");
    LogAllOpenQueues(r->server);

    if ((rc = ParseURI(r)) != OK)
    {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "error detected parsing URI");
        EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_URI_PARSE));
        rc = HTTP_BAD_REQUEST;
    }
    else
    {
        /* Get Appd specific info */
        sContext.psAppdContext = GetAppdContext();
        LogAllAppdContexts(r->server);

        /* switch based on req/action  (sContext.szArgReq/sContext.szArgAction) */
        if (strcmp(sContext.szArgReq, "CSI")==0)
        {
            if (strcmp(sContext.szArgAction, "CALL")==0)
            {
                rc = CsiCall(r);
            }
            else
            {
                ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid action <%s> for req <%s>", sContext.szArgAction, sContext.szArgReq);
                EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INVALID_ACTION, sContext.szArgAction, sContext.szArgReq));
                rc = HTTP_BAD_REQUEST;
            }
        }
        else if (strcmp(sContext.szArgReq, "ADMIN")==0)
        {
            if ((strcmp(sContext.szArgAction, "CHECK"  )==0)
             || (strcmp(sContext.szArgAction, "START"  )==0)
             || (strcmp(sContext.szArgAction, "STOP"   )==0)
             || (strcmp(sContext.szArgAction, "RESTART")==0))
            {
                rc = AdminRequest(r);
            }
            else
            {
                ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid action <%s> for req <%s>", sContext.szArgAction, sContext.szArgReq);
                EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INVALID_ACTION, sContext.szArgAction, sContext.szArgReq));
                rc = HTTP_BAD_REQUEST;
            }
        }
        else
        {
            ap_log_rerror(APLOG_MARK, APLOG_WARNING, r, "invalid req <%s>", sContext.szArgReq);
            EVR_TRACELOG(EVR_CRITICAL, (H2C_MSG_INVALID_REQ, sContext.szArgReq));
            rc = HTTP_BAD_REQUEST;
        }
    }

    h2c_rdebug(r, "Http2csiHandler END");
    LogAllOpenQueues(r->server);
    h2c_rdebug (r, "Http2csiHandler returns %d", rc);
    return rc;
}
