/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  RAPMGMT 
*
* @FILE     rapmgmt_tux.c

* @ABSTRACT
* Apache RAP ManaGeMenT (RAPMGMT) : C program providing the Tuxedo interface
*
* This file contains:
* the functions that implement the Tuxedo calls
*
* @INCLUDE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

/* TUXEDO include files */
#include <atmi.h>
#include <fml32.h>
#include <tpadm.h>

#include "rapmgmt_prcon.h"

extern int Ferror32;
extern int tperrno;


int main (void)
{
    char szRapName[48];
    char c;
    ssize_t n;
    size_t index;
    FBFR32 *ibuf;
    FBFR32 *obuf;
    int iTuxError;
    long lLen;
    long lBufLength = 10240;

    if (tpinit((TPINIT *) NULL) == -1)
    {
        exit(1);
    }

    /* allocate the buffers */
    if ( (ibuf = (FBFR32 *) tpalloc("FML32", 0 , lBufLength)) == NULL)
    {
        iTuxError = tperrno;
        tpterm();
        exit(2);
    }
    if ( (obuf = (FBFR32 *) tpalloc("FML32", 0 , lBufLength)) == NULL)
    {
        iTuxError = tperrno;
        tpfree((char *) ibuf);
        tpterm();
        exit(3);
    }

    index = 0; 
    memset(szRapName, '\0', sizeof(szRapName));
    while ((n = read(STDIN_FILENO, &c, 1)) > 0)
    {
        if (c == '\n')
        {
            /* we got a complete rap name */

            /* Clear the buffers */
            if (Finit32((FBFR32 *) ibuf, Fsizeof32(ibuf)) < 0)
            {
                iTuxError = Ferror32;
                tpfree((char *) ibuf);
                tpfree((char *) obuf);
                tpterm();
                exit(4);
            }
            if (Finit32(obuf, Fsizeof32(obuf)) < 0)
            {
                iTuxError = Ferror32;
                tpfree((char *) ibuf);
                tpfree((char *) obuf);
                tpterm();
                exit(5);
            }
            /* Set MIB(5) attributes defining request type */
            if (   Fadd32(ibuf, TA_OPERATION, TUXEDO_GET, 0) < 0
                || Fadd32(ibuf, TA_CLASS, TUXEDO_TSERVER, 0) < 0
                || Fadd32(ibuf, TA_SERVERNAME, szRapName, 0) < 0
               )
            {
                iTuxError = Ferror32;
                tpfree((char *) ibuf);
                tpfree((char *) obuf);
                tpterm();
                exit(6);
            }

            if (0 > tpcall(TUXEDO_TMIB, (char *) ibuf, 0, (char **) &obuf, &lLen, 0))
            {
                iTuxError = tperrno;
                tpfree((char *) ibuf);
                tpfree((char *) obuf);
                tpterm();
                exit(7);
            }
            else
            {
                char *ta_state;
                int length = 0;
     
                ta_state = Fvals32(obuf, TA_STATE, 0);
                length = strlen(ta_state);
                if (write(STDOUT_FILENO, ta_state, length) != length)
                {
                    tpfree((char *) ibuf);
                    tpfree((char *) obuf);
                    tpterm();
                    exit(8);
                }
                if (write(STDOUT_FILENO, "\n", 1) != 1)
                {
                    tpfree((char *) ibuf);
                    tpfree((char *) obuf);
                    tpterm();
                    exit(8);
                }
            }

            /* Resetting fields for a new rap */
            index = 0;
            memset(szRapName, '\0', sizeof(szRapName));
        }
        else
            szRapName[index++] = c;
    }
     
    tpfree((char *) ibuf);
    tpfree((char *) obuf);
    tpterm();

    if (n < 0)
    {
        /* error during 'read', O meaning EOF */
        exit(9);
    }

    exit (0);
}
