/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  APACHE
*
* @FILE     proxy_connect.c
*
* @ABSTRACT
* Modified Apache source
*
* This file contains:
*
* @INCLUDE


/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2002 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/* CONNECT method for Apache proxy */

#include "mod_proxy.h"
#include "http_log.h"
#include "http_main.h"

#ifdef HAVE_BSTRING_H
#include <bstring.h>            /* for IRIX, FD_SET calls bzero() */
#endif

#include <sys/time.h>		/* REX: for timestamps between gateways */
/*
 * This handles Netscape CONNECT method secure proxy requests.
 * A connection is opened to the specified host and data is
 * passed through between the WWW site and the browser.
 *
 * This code is based on the INTERNET-DRAFT document
 * "Tunneling SSL Through a WWW Proxy" currently at
 * http://www.mcom.com/newsref/std/tunneling_ssl.html.
 *
 * If proxyhost and proxyport are set, we send a CONNECT to
 * the specified proxy..
 *
 * FIXME: this is bad, because it does its own socket I/O
 *        instead of using the I/O in buff.c.  However,
 *        the I/O in buff.c blocks on reads, and because
 *        this function doesn't know how much data will
 *        be sent either way (or when) it can't use blocking
 *        I/O.  This may be very implementation-specific
 *        (to Linux).  Any suggestions?
 * FIXME: this doesn't log the number of bytes sent, but
 *        that may be okay, since the data is supposed to
 *        be transparent. In fact, this doesn't log at all
 *        yet. 8^)
 * FIXME: doesn't check any headers initally sent from the
 *        client.
 * FIXME: should allow authentication, but hopefully the
 *        generic proxy authentication is good enough.
 * FIXME: no check for r->assbackwards, whatever that is.
 */

static int allowed_port(proxy_server_conf *conf, int port)
{
    int i;
    int *list = (int *)conf->allowed_connect_ports->elts;

    for (i = 0; i < conf->allowed_connect_ports->nelts; i++) {
        if (port == list[i])
            return 1;
    }
    return 0;
}

/**********************************************
* ccn_getline
* will read a single line from *in and copy 
* it in *out, l limits the size to be copied
* it will return -1 when the headers are passed
**********************************************/
int ccn_getline(char *in, char *out, int l)
{
    char *pos;
    int retval = strlen(in);
    int n = 0;
    int total = 0;

    pos = in;

    if (retval <= 0) {
        total = -1;
        return CCN_EMPTY_LINE;
    }

    while (n < retval && *pos != '\n')
    {
        ++pos;
        ++total;
        ++n;
    }

    /* an empty line means end of headers
       see rfc2616 chap 4.1               */
    if (n <= 1 && *pos == '\n' )
    {
        total = -2;
        return CCN_END_OF_HEADER;
    }
		
    ++total;

    while (pos > (in + 1) && (*(pos - 1) == ' ' || *(pos - 1) == '\t')) {
       --pos;
       --n;
    }

    if (n > l)
        n = l;

    strncpy(out,in,n);
    *(out+n) = '\0';

    return total;
}





/**********************************************
* ccn_full_header_recv
* will read a buffer to see if we received a 
* full header, returns 1 if so, 0 if not.
**********************************************/
int ccn_full_header_recv(char *in, int l)
{
    int len;
    int readlen = 0;
    char cline[MAX_STRING_LEN];
 
    /* loop through the buffer */ 
    while ((len = ccn_getline(in+readlen, cline, MAX_STRING_LEN)) > 0) {
      readlen += len; 
    }

   if (len == CCN_END_OF_HEADER)
      return CCN_FULL_HEADER; 

   return CCN_PART_HEADER;
}

/**********************************************
* ccn_proxy_read_response_line
* scans the buffer for the HTTP return and reason
* code. Expects a single line.
* result will be saved in the request_rec structure r
**********************************************/
int ccn_proxy_read_response_line(request_rec *r, char *buffer) 
{
    long len;
    int major;
    int minor;

    if (2 != sscanf(buffer, "HTTP/%u.%u", &major, &minor)) {
        /* if no response, default to HTTP/1.1 - is this correct? */
        major = 1;
        minor = 1;
    }
         
    /* If not an HTTP/1 message */
    if (major < 1) {
        return HTTP_BAD_GATEWAY;
    }         
                  
    /* there need not be a reason phrase in the response,
     * and ap_getline() already deleted trailing whitespace.
     * But RFC2616 requires a SP after the Status-Code. Add one:
     */       

    if (strlen(buffer) < sizeof("HTTP/1.x 200 ")-1)
       buffer = ap_pstrcat(r->pool, buffer, " ", NULL);
    buffer[12] = '\0';

    if(ap_is_HTTP_SERVER_ERROR(atoi(&buffer[9])))
    {
       r->status = atoi(&buffer[9]);
       buffer[12] = ' ';
       r->status_line = ap_pstrdup(r->pool, &buffer[9]);
    }

    return OK;
}            

/**********************************************
* ccn_proxy_read_headers
* scans the buffer for both the http return code
* and all headers within the buffer.
* returns the table will all headers found.
**********************************************/
table *ccn_proxy_read_headers(request_rec *r, char *buffer)
{
    table *resp_hdrs;
    int len;
    int readlen = 0;
    char *value, *end;
    char cline[MAX_STRING_LEN];

    resp_hdrs = ap_make_table(r->pool, 20);

    while ((len = ccn_getline(buffer+readlen, cline, MAX_STRING_LEN)) > 0) {

        readlen += len;

        if (!(value = strchr(cline, ':'))) {   /* Find the colon separator */

            /*
             *. Buggy MS IIS servers sometimes return invalid headers (an
             * extra "HTTP/1.0 200, OK" line sprinkled in between the usual
             * MIME headers). Try to deal with it in a sensible way, but log
             * the fact. XXX: The mask check is buggy if we ever see an
             * HTTP/1.10
             */         
               
            if (!ap_checkmask(cline, "HTTP/#.# ###*")) {
                return NULL;
            }
            else
            {
                ccn_proxy_read_response_line(r, cline);
            }
             
            ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, r->server,
                         "proxy: Ignoring duplicate HTTP status line "
                         "returned by buggy server %s (%s)", r->uri, r->method);
            continue;
        }
    
        *value = '\0';
        ++value;

        /*
         * XXX: RFC2068 defines only SP and HT as whitespace, this test is
         * wrong... and so are many others probably.
         */
        while (ap_isspace(*value))
            ++value;
    
        for (end = &value[strlen(value) - 1]; end > value && ap_isspace(*end); --end)
            *end = '\0';
    
        ap_table_add(resp_hdrs, cline, value);
    }
    return resp_hdrs;
}        


int ap_proxy_connect_handler(request_rec *r, cache_req *c, char *url,
                                 const char *proxyhost, int proxyport)
{
    struct sockaddr_in server;
    struct in_addr destaddr;
    struct hostent server_hp;
    const char *host, *err;
    char *p;
    int port, sock;
    char buffer[HUGE_STRING_LEN];
    int nbytes, i, j;
    fd_set fds;

    /* OR10237  */
    double t3,t4;	    /* times */
    int remproxy = 0;       /* flag to hold amount of gateways to pass */
    int firstpass = -1;     /* flag to hold amount of gateways to pass */
    struct timeval tim;     /* milliseconds code */
    char ctemp[20] = {""};  /* holder for the timestamp when added to the header */
    char *temp_buffer;
    table *temp_hdrs;
    pool *temp_pool;
    int lLength = MAX_STRING_LEN * 22;
    int nbufbytes = 0;
    /*  */

    t3 = t4 = 0;

    /* initialise the pool and our temp_buffer */
    temp_pool = ap_make_sub_pool(r->pool);
    temp_buffer = (char *) ap_palloc(temp_pool, lLength);
    memset(temp_buffer, '\0', lLength);

    void *sconf = r->server->module_config;
    proxy_server_conf *conf =
    (proxy_server_conf *)ap_get_module_config(sconf, &proxy_module);
    struct noproxy_entry *npent = (struct noproxy_entry *) conf->noproxies->elts;

    memset(&server, '\0', sizeof(server));
    server.sin_family = AF_INET;

    /* Break the URL into host:port pairs */

    host = url;
    p = strchr(url, ':');
    if (p == NULL)
        port = DEFAULT_HTTPS_PORT;
    else {
        port = atoi(p + 1);
        *p = '\0';
    }

/* check if ProxyBlock directive on this host */
    destaddr.s_addr = ap_inet_addr(host);
    for (i = 0; i < conf->noproxies->nelts; i++) {
        if ((npent[i].name != NULL && strstr(host, npent[i].name) != NULL)
            || destaddr.s_addr == npent[i].addr.s_addr
            || npent[i].name[0] == '*')
            return ap_proxyerror(r, HTTP_FORBIDDEN,
                                 "Connect to remote machine blocked");
    }

    /* Check if it is an allowed port */
    if (conf->allowed_connect_ports->nelts == 0) {
        /* Default setting if not overridden by AllowCONNECT */
        switch (port) {
        case DEFAULT_HTTPS_PORT:
        case DEFAULT_SNEWS_PORT:
            break;
        default:
            return HTTP_FORBIDDEN;
        }
    }
    else if (!allowed_port(conf, port))
        return HTTP_FORBIDDEN;

    if (proxyhost) {
        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
             "CONNECT to remote proxy %s on port %d", proxyhost, proxyport);
    }
    else {
        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                     "CONNECT to %s on port %d", host, port);
    }

    /* Nasty cast to work around broken terniary expressions on MSVC */
    server.sin_port = htons((unsigned short)(proxyport ? proxyport : port));
    err = ap_proxy_host2addr(proxyhost ? proxyhost : host, &server_hp);

    if (err != NULL)
        return ap_proxyerror(r,
            proxyhost ? HTTP_BAD_GATEWAY : HTTP_INTERNAL_SERVER_ERROR, err);

    sock = ap_psocket(r->pool, PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (sock == -1) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, r, "proxy: error creating socket");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

#ifdef CHECK_FD_SETSIZE
    if (sock >= FD_SETSIZE) {
        ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_WARNING, NULL,
                     "proxy_connect_handler: filedescriptor (%u) "
                     "larger than FD_SETSIZE (%u) "
                     "found, you probably need to rebuild Apache with a "
                     "larger FD_SETSIZE", sock, FD_SETSIZE);
        ap_pclosesocket(r->pool, sock);
        return HTTP_INTERNAL_SERVER_ERROR;
    }
#endif

    j = 0;
    while (server_hp.h_addr_list[j] != NULL) {
        memcpy(&server.sin_addr, server_hp.h_addr_list[j],
               sizeof(struct in_addr));
        i = ap_proxy_doconnect(sock, &server, r);
        if (i == 0)
            break;
        j++;
    }
    if (i == -1) {
        ap_pclosesocket(r->pool, sock);
        return ap_proxyerror(r, HTTP_INTERNAL_SERVER_ERROR, ap_pstrcat(r->pool,
        "Could not connect to remote machine:<br>", strerror(errno), NULL));
    }

    /*
     * If we are connecting through a remote proxy, we need to pass the
     * CONNECT request on to it.
     */
    if (proxyport) {
        /*
         * FIXME: We should not be calling write() directly, but we currently
         * have no alternative.  Error checking ignored.  Also, we force a
         * HTTP/1.0 request to keep things simple.
         */
        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                     "Sending the CONNECT request to the remote proxy");
        ap_snprintf(buffer, sizeof(buffer), "CONNECT %s HTTP/1.0" CRLF, r->uri);
        send(sock, buffer, strlen(buffer), 0);

        /* CCN
         * passing to a second gateway
         */
        ap_snprintf(buffer, sizeof(buffer), "t1: %s" CRLF,ap_table_get(r->notes,"t1"));
        send(sock, buffer, strlen(buffer), 0);
        ap_snprintf(buffer, sizeof(buffer), "ccnproxy: %s" CRLF,"1");
        send(sock, buffer, strlen(buffer), 0);
        remproxy = 1; 
        /* /CCN */

	ap_snprintf(buffer, sizeof(buffer),
                    "Proxy-agent: %s" CRLF CRLF, ap_get_server_version());
        send(sock, buffer, strlen(buffer), 0);
    }
    else {
        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                     "Returning 200 OK Status");
        r->bytes_sent += ap_rvputs(r, "HTTP/1.0 200 Connection established" CRLF, NULL);
 
        gettimeofday(&tim, NULL);
        t3=tim.tv_sec+(tim.tv_usec/1000000.0);
        sprintf(ctemp,"%f",t3);
        ap_table_set(r->notes, "t3",ctemp);

        if (ap_table_get(r->headers_in, "ccnproxy") != NULL)
        {
           r->bytes_sent += ap_rvputs(r, "t2: ", ap_table_get(r->notes,"t2"), CRLF, NULL);
           r->bytes_sent += ap_rvputs(r, "t3: ", ap_table_get(r->notes,"t3"), CRLF, NULL);
        }
 
        r->bytes_sent += ap_rvputs(r, "Proxy-agent: ", ap_get_server_version(), CRLF CRLF, NULL);
 
        ap_bflush(r->connection->client);
    }

    while (1) {                 /* Infinite loop until error (one side closes
                                 * the connection) */
        FD_ZERO(&fds);
        FD_SET(sock, &fds);
        FD_SET(ap_bfileno(r->connection->client, B_WR), &fds);

        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                     "Going to sleep (select)");
        i = ap_select((ap_bfileno(r->connection->client, B_WR) > sock ?
                       ap_bfileno(r->connection->client, B_WR) + 1 :
                       sock + 1), &fds, NULL, NULL, NULL);
        ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                     "Woke from select(), i=%d", i);

        if (i) {
            if (FD_ISSET(sock, &fds)) {
                ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                             "sock was set");
                if ((nbytes = recv(sock, buffer, HUGE_STRING_LEN, 0)) != 0) {
                    /* CCN */
		    if( remproxy == 1 && firstpass == -1 )
                    {
                       /* we will set firstpass to 0 to parse the buffer once we
                        * have fully received the header (empty line rule)
                        */
                        if (nbytes > 0)
                        {  
                           /* add the received bytes to the buffer */
                           strncpy(temp_buffer+nbufbytes, buffer, nbytes); 
                           nbufbytes += nbytes;
                           
                           /* if we receive a full header we will parse it for timestamps */

                           if ( CCN_FULL_HEADER == ccn_full_header_recv(temp_buffer,nbufbytes))
                           {
                              firstpass = 0;
                           }
                        }    
                    }
                    
                    if( remproxy == 1 && firstpass == 0 )	
		    { 
                       /* we only want to parse the first buffer being sent
                        * this buffer can contain the HTTP 200 OK to create the tunnel
                        */ 
                       firstpass = 1;
                      
                       /* read the buffer in a non destructive way
                        * if we want to connect to an external webserver (outside ccn)
                        * and the connection fails the status will be set to 500 instead of
                        * 200
                        */ 
                       temp_hdrs = ccn_proxy_read_headers(r, temp_buffer);
                       
                       if(ap_table_get(temp_hdrs, "t2") != NULL)
                          ap_table_set(r->notes, "t2",ap_table_get(temp_hdrs, "t2"));
                       if(ap_table_get(temp_hdrs, "t3") != NULL)
                          ap_table_set(r->notes, "t3",ap_table_get(temp_hdrs, "t3"));

                        /* at this time our tunnel is either up or will
                         * never be up, t4 is taken either way.
                         */
                        gettimeofday(&tim, NULL);
                        t4=tim.tv_sec+(tim.tv_usec/1000000.0);
                        sprintf(ctemp,"%f",t4);
                        ap_table_set(r->notes, "t4",ctemp);

                        /* we can clean up our pool */
                        ap_destroy_pool(temp_pool);
		    } 
                    /* CCN */    
                    if (nbytes == -1)
                        break;
                    if (send(ap_bfileno(r->connection->client, B_WR), buffer,
                             nbytes, 0) == EOF)
                        break;
                    r->bytes_sent += nbytes;
                    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO,
                             r->server, "Wrote %d bytes to client", nbytes);
                }
                else
                    break;
            }
            else if (FD_ISSET(ap_bfileno(r->connection->client, B_WR), &fds)) {
                ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r->server,
                             "client->fd was set");
                if ((nbytes = recv(ap_bfileno(r->connection->client, B_WR),
                                   buffer, HUGE_STRING_LEN, 0)) != 0) {
                    if (nbytes == -1)
                        break;
                    r->read_length += nbytes;
                    if (send(sock, buffer, nbytes, 0) == EOF)
                        break;
                    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO,
                             r->server, "Wrote %d bytes to server", nbytes);
                }
                else
                    break;
            }
            else
                break;          /* Must be done waiting */
        }
        else
            break;
    }
    
    ap_pclosesocket(r->pool, sock);

    return OK;
}
